# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.model import hashutil
from swh.web.common import converters


def test_fmap():
    assert [2, 3, None, 4] == converters.fmap(lambda x: x+1, [1, 2, None, 3])
    assert [11, 12, 13] == list(converters.fmap(lambda x: x+10,
                                                map(lambda x: x, [1, 2, 3])))
    assert {'a': 2, 'b': 4} == converters.fmap(lambda x: x*2, {'a': 1, 'b': 2})
    assert 100 == converters.fmap(lambda x: x*10, 10)
    assert {'a': [2, 6], 'b': 4} == converters.fmap(lambda x: x*2,
                                                    {'a': [1, 3], 'b': 2})
    assert converters.fmap(lambda x: x, None) is None


def test_from_swh():
    some_input = {
        'a': 'something',
        'b': 'someone',
        'c': b'sharp-0.3.4.tgz',
        'd': hashutil.hash_to_bytes(
            'b04caf10e9535160d90e874b45aa426de762f19f'),
        'e': b'sharp.html/doc_002dS_005fISREG.html',
        'g': [b'utf-8-to-decode', b'another-one'],
        'h': 'something filtered',
        'i': {'e': b'something'},
        'j': {
            'k': {
                'l': [b'bytes thing', b'another thingy', b''],
                'n': 'dont care either'
            },
            'm': 'dont care'
        },
        'o': 'something',
        'p': b'foo',
        'q': {'extra-headers': [['a', b'intact']]},
        'w': None,
        'r': {'p': 'also intact',
              'q': 'bar'},
        's': {
            'timestamp': 42,
            'offset': -420,
            'negative_utc': None,
        },
        's1': {
            'timestamp': {'seconds': 42, 'microseconds': 0},
            'offset': -420,
            'negative_utc': None,
        },
        's2': datetime.datetime(
            2013, 7, 1, 20, 0, 0,
            tzinfo=datetime.timezone.utc),
        't': None,
        'u': None,
        'v': None,
        'x': None,
    }

    expected_output = {
        'a': 'something',
        'b': 'someone',
        'c': 'sharp-0.3.4.tgz',
        'd': 'b04caf10e9535160d90e874b45aa426de762f19f',
        'e': 'sharp.html/doc_002dS_005fISREG.html',
        'g': ['utf-8-to-decode', 'another-one'],
        'i': {'e': 'something'},
        'j': {
            'k': {
                'l': ['bytes thing', 'another thingy', '']
            }
        },
        'p': 'foo',
        'q': {'extra-headers': [['a', 'intact']]},
        'w': {},
        'r': {'p': 'also intact',
              'q': 'bar'},
        's': '1969-12-31T17:00:42-07:00',
        's1': '1969-12-31T17:00:42-07:00',
        's2': '2013-07-01T20:00:00+00:00',
        'u': {},
        'v': [],
        'x': None,
    }

    actual_output = converters.from_swh(
        some_input,
        hashess={'d', 'o', 'x'},
        bytess={'c', 'e', 'g', 'l'},
        dates={'s', 's1', 's2'},
        blacklist={'h', 'm', 'n', 'o'},
        removables_if_empty={'t'},
        empty_dict={'u'},
        empty_list={'v'},
        convert={'p', 'q', 'w'},
        convert_fn=converters.convert_revision_metadata)

    assert expected_output == actual_output


def test_from_swh_edge_cases_do_no_conversion_if_none_or_not_bytes():
    some_input = {
        'a': 'something',
        'b': None,
        'c': 'someone',
        'd': None,
        'e': None
    }

    expected_output = {
        'a': 'something',
        'b': None,
        'c': 'someone',
        'd': None,
        'e': None
    }

    actual_output = converters.from_swh(some_input,
                                        hashess={'a', 'b'},
                                        bytess={'c', 'd'},
                                        dates={'e'})

    assert expected_output == actual_output


def test_from_swh_edge_cases_convert_invalid_utf8_bytes():
    some_input = {
        'a': 'something',
        'b': 'someone',
        'c': b'a name \xff',
        'd': b'an email \xff',
    }

    expected_output = {
        'a': 'something',
        'b': 'someone',
        'c': 'a name \\xff',
        'd': 'an email \\xff',
        'decoding_failures': ['c', 'd']
    }

    actual_output = converters.from_swh(some_input,
                                        hashess={'a', 'b'},
                                        bytess={'c', 'd'})
    for v in ['a', 'b', 'c', 'd']:
        assert expected_output[v] == actual_output[v]
    assert (len(expected_output['decoding_failures']) ==
            len(actual_output['decoding_failures']))
    for v in expected_output['decoding_failures']:
        assert v in actual_output['decoding_failures']


def test_from_swh_empty():
    assert {} == converters.from_swh({})


def test_from_swh_none():
    assert converters.from_swh(None) is None


def test_from_origin():
    origin_input = {
        'id': 9,
        'type': 'ftp',
        'url': 'rsync://ftp.gnu.org/gnu/octave',
    }

    expected_origin = {
        'id': 9,
        'type': 'ftp',
        'url': 'rsync://ftp.gnu.org/gnu/octave',
    }

    actual_origin = converters.from_origin(origin_input)

    assert actual_origin == expected_origin


def test_from_origin_visit():
    snap_hash = 'b5f0b7f716735ebffe38505c60145c4fd9da6ca3'

    for snap in [snap_hash, None]:
        visit = {
            'date': {
                'timestamp': datetime.datetime(
                    2015, 1, 1, 22, 0, 0,
                    tzinfo=datetime.timezone.utc).timestamp(),
                'offset': 0,
                'negative_utc': False,
            },
            'origin': 10,
            'visit': 100,
            'metadata': None,
            'status': 'full',
            'snapshot': hashutil.hash_to_bytes(snap) if snap else snap,
        }

        expected_visit = {
            'date': '2015-01-01T22:00:00+00:00',
            'origin': 10,
            'visit': 100,
            'metadata': {},
            'status': 'full',
            'snapshot': snap_hash if snap else snap
        }

        actual_visit = converters.from_origin_visit(visit)

        assert actual_visit == expected_visit


def test_from_release():
    release_input = {
        'id': hashutil.hash_to_bytes(
            'aad23fa492a0c5fed0708a6703be875448c86884'),
        'target': hashutil.hash_to_bytes(
            '5e46d564378afc44b31bb89f99d5675195fbdf67'),
        'target_type': 'revision',
        'date': {
            'timestamp': datetime.datetime(
                2015, 1, 1, 22, 0, 0,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'author': {
            'name': b'author name',
            'fullname': b'Author Name author@email',
            'email': b'author@email',
        },
        'name': b'v0.0.1',
        'message': b'some comment on release',
        'synthetic': True,
    }

    expected_release = {
        'id': 'aad23fa492a0c5fed0708a6703be875448c86884',
        'target': '5e46d564378afc44b31bb89f99d5675195fbdf67',
        'target_type': 'revision',
        'date': '2015-01-01T22:00:00+00:00',
        'author': {
            'name': 'author name',
            'fullname': 'Author Name author@email',
            'email': 'author@email',
        },
        'name': 'v0.0.1',
        'message': 'some comment on release',
        'target_type': 'revision',
        'synthetic': True,
    }

    actual_release = converters.from_release(release_input)

    assert actual_release == expected_release


def test_from_release_no_revision():
    release_input = {
        'id': hashutil.hash_to_bytes(
            'b2171ee2bdf119cd99a7ec7eff32fa8013ef9a4e'),
        'target': None,
        'date': {
            'timestamp': datetime.datetime(
                2016, 3, 2, 10, 0, 0,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': True,

        },
        'name': b'v0.1.1',
        'message': b'comment on release',
        'synthetic': False,
        'author': {
            'name': b'bob',
            'fullname': b'Bob bob@alice.net',
            'email': b'bob@alice.net',
        },
    }

    expected_release = {
        'id': 'b2171ee2bdf119cd99a7ec7eff32fa8013ef9a4e',
        'target': None,
        'date': '2016-03-02T10:00:00-00:00',
        'name': 'v0.1.1',
        'message': 'comment on release',
        'synthetic': False,
        'author': {
            'name': 'bob',
            'fullname': 'Bob bob@alice.net',
            'email': 'bob@alice.net',
        },
    }

    actual_release = converters.from_release(release_input)

    assert actual_release == expected_release


def test_from_revision():
    revision_input = {
        'id': hashutil.hash_to_bytes(
            '18d8be353ed3480476f032475e7c233eff7371d5'),
        'directory': hashutil.hash_to_bytes(
            '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6'),
        'author': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'committer': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'message': b'synthetic revision message',
        'date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'committer_date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'synthetic': True,
        'type': 'tar',
        'parents': [
            hashutil.hash_to_bytes(
                '29d8be353ed3480476f032475e7c244eff7371d5'),
            hashutil.hash_to_bytes(
                '30d8be353ed3480476f032475e7c244eff7371d5')
        ],
        'children': [
            hashutil.hash_to_bytes(
                '123546353ed3480476f032475e7c244eff7371d5'),
        ],
        'metadata': {
            'extra_headers': [['gpgsig', b'some-signature']],
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912',

            }]
        },
    }

    expected_revision = {
        'id': '18d8be353ed3480476f032475e7c233eff7371d5',
        'directory': '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6',
        'author': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'committer': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'message': 'synthetic revision message',
        'date': "2000-01-17T11:23:54+00:00",
        'committer_date': "2000-01-17T11:23:54+00:00",
        'children': [
            '123546353ed3480476f032475e7c244eff7371d5'
        ],
        'parents': [
            '29d8be353ed3480476f032475e7c244eff7371d5',
            '30d8be353ed3480476f032475e7c244eff7371d5'
        ],
        'type': 'tar',
        'synthetic': True,
        'metadata': {
            'extra_headers': [['gpgsig', 'some-signature']],
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912'
            }]
        },
        'merge': True
    }

    actual_revision = converters.from_revision(revision_input)

    assert actual_revision == expected_revision


def test_from_revision_nomerge():
    revision_input = {
        'id': hashutil.hash_to_bytes(
            '18d8be353ed3480476f032475e7c233eff7371d5'),
        'parents': [
            hashutil.hash_to_bytes(
                '29d8be353ed3480476f032475e7c244eff7371d5')
        ]
    }

    expected_revision = {
        'id': '18d8be353ed3480476f032475e7c233eff7371d5',
        'parents': [
            '29d8be353ed3480476f032475e7c244eff7371d5'
        ],
        'merge': False
    }

    actual_revision = converters.from_revision(revision_input)

    assert actual_revision == expected_revision


def test_from_revision_noparents():
    revision_input = {
        'id': hashutil.hash_to_bytes(
            '18d8be353ed3480476f032475e7c233eff7371d5'),
        'directory': hashutil.hash_to_bytes(
            '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6'),
        'author': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'committer': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'message': b'synthetic revision message',
        'date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'committer_date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'synthetic': True,
        'type': 'tar',
        'children': [
            hashutil.hash_to_bytes(
                '123546353ed3480476f032475e7c244eff7371d5'),
        ],
        'metadata': {
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912',

            }]
        },
    }

    expected_revision = {
        'id': '18d8be353ed3480476f032475e7c233eff7371d5',
        'directory': '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6',
        'author': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'committer': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'message': 'synthetic revision message',
        'date': "2000-01-17T11:23:54+00:00",
        'committer_date': "2000-01-17T11:23:54+00:00",
        'children': [
            '123546353ed3480476f032475e7c244eff7371d5'
        ],
        'type': 'tar',
        'synthetic': True,
        'metadata': {
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912'
            }]
        }
    }

    actual_revision = converters.from_revision(revision_input)

    assert actual_revision == expected_revision


def test_from_revision_invalid():
    revision_input = {
        'id': hashutil.hash_to_bytes(
            '18d8be353ed3480476f032475e7c233eff7371d5'),
        'directory': hashutil.hash_to_bytes(
            '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6'),
        'author': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'committer': {
            'name': b'Software Heritage',
            'fullname': b'robot robot@softwareheritage.org',
            'email': b'robot@softwareheritage.org',
        },
        'message': b'invalid message \xff',
        'date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'committer_date': {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc).timestamp(),
            'offset': 0,
            'negative_utc': False,
        },
        'synthetic': True,
        'type': 'tar',
        'parents': [
            hashutil.hash_to_bytes(
                '29d8be353ed3480476f032475e7c244eff7371d5'),
            hashutil.hash_to_bytes(
                '30d8be353ed3480476f032475e7c244eff7371d5')
        ],
        'children': [
            hashutil.hash_to_bytes(
                '123546353ed3480476f032475e7c244eff7371d5'),
        ],
        'metadata': {
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912',

            }]
        },
    }

    expected_revision = {
        'id': '18d8be353ed3480476f032475e7c233eff7371d5',
        'directory': '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6',
        'author': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'committer': {
            'name': 'Software Heritage',
            'fullname': 'robot robot@softwareheritage.org',
            'email': 'robot@softwareheritage.org',
        },
        'message': None,
        'message_decoding_failed': True,
        'date': "2000-01-17T11:23:54+00:00",
        'committer_date': "2000-01-17T11:23:54+00:00",
        'children': [
            '123546353ed3480476f032475e7c244eff7371d5'
        ],
        'parents': [
            '29d8be353ed3480476f032475e7c244eff7371d5',
            '30d8be353ed3480476f032475e7c244eff7371d5'
        ],
        'type': 'tar',
        'synthetic': True,
        'metadata': {
            'original_artifact': [{
                'archive_type': 'tar',
                'name': 'webbase-5.7.0.tar.gz',
                'sha1': '147f73f369733d088b7a6fa9c4e0273dcd3c7ccd',
                'sha1_git': '6a15ea8b881069adedf11feceec35588f2cfe8f1',
                'sha256': '401d0df797110bea805d358b85bcc1ced29549d3d73f'
                '309d36484e7edf7bb912'
            }]
        },
        'merge': True
    }

    actual_revision = converters.from_revision(revision_input)

    assert actual_revision == expected_revision


def test_from_content_none():
    assert converters.from_content(None) is None


def test_from_content():
    content_input = {
        'sha1': hashutil.hash_to_bytes(
            '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5'),
        'sha256': hashutil.hash_to_bytes(
            '39007420ca5de7cb3cfc15196335507e'
            'e76c98930e7e0afa4d2747d3bf96c926'),
        'blake2s256': hashutil.hash_to_bytes(
            '49007420ca5de7cb3cfc15196335507e'
            'e76c98930e7e0afa4d2747d3bf96c926'),
        'sha1_git': hashutil.hash_to_bytes(
            '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'),
        'ctime': 'something-which-is-filtered-out',
        'data': b'data in bytes',
        'length': 10,
        'status': 'hidden',
    }

    # 'status' is filtered
    expected_content = {
        'checksums': {
            'sha1': '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5',
            'sha256': '39007420ca5de7cb3cfc15196335507ee76c98'
                      '930e7e0afa4d2747d3bf96c926',
            'blake2s256': '49007420ca5de7cb3cfc15196335507ee7'
                          '6c98930e7e0afa4d2747d3bf96c926',
            'sha1_git': '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03',
        },
        'data': b'data in bytes',
        'length': 10,
        'status': 'absent',
    }

    actual_content = converters.from_content(content_input)

    assert actual_content == expected_content


def test_from_person():
    person_input = {
        'id': 10,
        'anything': 'else',
        'name': b'bob',
        'fullname': b'bob bob@alice.net',
        'email': b'bob@foo.alice',
    }

    expected_person = {
        'id': 10,
        'anything': 'else',
        'name': 'bob',
        'fullname': 'bob bob@alice.net',
        'email': 'bob@foo.alice',
    }

    actual_person = converters.from_person(person_input)

    assert actual_person == expected_person


def test_from_directory_entries():
    dir_entries_input = {
        'sha1': hashutil.hash_to_bytes(
            '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5'),
        'sha256': hashutil.hash_to_bytes(
            '39007420ca5de7cb3cfc15196335507e'
            'e76c98930e7e0afa4d2747d3bf96c926'),
        'sha1_git': hashutil.hash_to_bytes(
            '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'),
        'blake2s256': hashutil.hash_to_bytes(
            '685395c5dc57cada459364f0946d3dd45bad5fcbab'
            'c1048edb44380f1d31d0aa'),
        'target': hashutil.hash_to_bytes(
            '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'),
        'dir_id': hashutil.hash_to_bytes(
            '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'),
        'name': b'bob',
        'type': 10,
        'status': 'hidden',
    }

    expected_dir_entries = {
        'checksums': {
            'sha1': '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5',
            'sha256': '39007420ca5de7cb3cfc15196335507ee76c98'
                      '930e7e0afa4d2747d3bf96c926',
            'sha1_git': '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03',
            'blake2s256': '685395c5dc57cada459364f0946d3dd45bad5f'
                          'cbabc1048edb44380f1d31d0aa',
        },
        'target': '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03',
        'dir_id': '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03',
        'name': 'bob',
        'type': 10,
        'status': 'absent',
    }

    actual_dir_entries = converters.from_directory_entry(dir_entries_input)

    assert actual_dir_entries == expected_dir_entries


def test_from_filetype():
    content_filetype = {
        'id': hashutil.hash_to_bytes(
            '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5'),
        'encoding': 'utf-8',
        'mimetype': 'text/plain',
    }

    expected_content_filetype = {
        'id': '5c6f0e2750f48fa0bd0c4cf5976ba0b9e02ebda5',
        'encoding': 'utf-8',
        'mimetype': 'text/plain',
    }

    actual_content_filetype = converters.from_filetype(content_filetype)

    assert actual_content_filetype == expected_content_filetype



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.utils.html import escape
from hypothesis import given

from swh.web.common.utils import (
    reverse, format_utc_iso_date, get_swh_persistent_id,
    parse_timestamp
)
from swh.web.tests.django_asserts import assert_contains, assert_template_used
from swh.web.tests.strategies import (
    origin, revision, unknown_revision, new_origin
)


@given(revision())
def test_revision_browse(client, archive_data, revision):
    url = reverse('browse-revision',
                  url_args={'sha1_git': revision})

    revision_data = archive_data.revision_get(revision)

    author_name = revision_data['author']['name']
    committer_name = revision_data['committer']['name']
    dir_id = revision_data['directory']

    directory_url = reverse('browse-directory',
                            url_args={'sha1_git': dir_id})

    history_url = reverse('browse-revision-log',
                          url_args={'sha1_git': revision})

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/revision.html')
    assert_contains(resp, author_name)
    assert_contains(resp, committer_name)
    assert_contains(resp, directory_url)
    assert_contains(resp, history_url)

    for parent in revision_data['parents']:
        parent_url = reverse('browse-revision',
                             url_args={'sha1_git': parent})
        assert_contains(resp, '<a href="%s">%s</a>' %
                        (parent_url, parent))

    author_date = revision_data['date']
    committer_date = revision_data['committer_date']

    message_lines = revision_data['message'].split('\n')

    assert_contains(resp, format_utc_iso_date(author_date))
    assert_contains(resp, format_utc_iso_date(committer_date))
    assert_contains(resp, escape(message_lines[0]))
    assert_contains(resp, escape('\n'.join(message_lines[1:])))


@given(origin())
def test_revision_origin_browse(client, archive_data, origin):
    snapshot = archive_data.snapshot_get_latest(origin['url'])
    revision = archive_data.snapshot_get_head(snapshot)
    revision_data = archive_data.revision_get(revision)
    dir_id = revision_data['directory']

    origin_revision_log_url = reverse('browse-origin-log',
                                        url_args={'origin_url': origin['url']},  # noqa
                                        query_params={'revision': revision})

    url = reverse('browse-revision',
                  url_args={'sha1_git': revision},
                  query_params={'origin': origin['url']})

    resp = client.get(url)

    assert_contains(resp, origin_revision_log_url)

    for parent in revision_data['parents']:
        parent_url = reverse('browse-revision',
                             url_args={'sha1_git': parent},
                             query_params={'origin': origin['url']})
        assert_contains(resp, '<a href="%s">%s</a>' %
                        (parent_url, parent))

    assert_contains(resp, 'vault-cook-directory')
    assert_contains(resp, 'vault-cook-revision')

    swh_rev_id = get_swh_persistent_id('revision', revision)
    swh_rev_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_rev_id})
    assert_contains(resp, swh_rev_id)
    assert_contains(resp, swh_rev_id_url)

    swh_dir_id = get_swh_persistent_id('directory', dir_id)
    swh_dir_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_dir_id})
    assert_contains(resp, swh_dir_id)
    assert_contains(resp, swh_dir_id_url)

    assert_contains(resp, 'swh-take-new-snapshot')


@given(revision())
def test_revision_log_browse(client, archive_data, revision):
    per_page = 10

    revision_log = archive_data.revision_log(revision)

    revision_log_sorted = \
        sorted(revision_log,
               key=lambda rev: -parse_timestamp(
                   rev['committer_date']).timestamp())

    url = reverse('browse-revision-log',
                  url_args={'sha1_git': revision},
                  query_params={'per_page': per_page})

    resp = client.get(url)

    next_page_url = reverse('browse-revision-log',
                            url_args={'sha1_git': revision},
                            query_params={'offset': per_page,
                                          'per_page': per_page})

    nb_log_entries = per_page
    if len(revision_log_sorted) < per_page:
        nb_log_entries = len(revision_log_sorted)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/revision-log.html')
    assert_contains(resp, '<tr class="swh-revision-log-entry',
                    count=nb_log_entries)
    assert_contains(resp, '<a class="page-link">Newer</a>')

    if len(revision_log_sorted) > per_page:
        assert_contains(resp, '<a class="page-link" href="%s">Older</a>' %  # noqa
                            escape(next_page_url))

    for log in revision_log_sorted[:per_page]:
        revision_url = reverse('browse-revision',
                               url_args={'sha1_git': log['id']})
        assert_contains(resp, log['id'][:7])
        assert_contains(resp, log['author']['name'])
        assert_contains(resp, format_utc_iso_date(log['date']))
        assert_contains(resp, escape(log['message']))
        assert_contains(resp, format_utc_iso_date(log['committer_date']))  # noqa
        assert_contains(resp, revision_url)

    if len(revision_log_sorted) <= per_page:
        return

    resp = client.get(next_page_url)

    prev_page_url = reverse('browse-revision-log',
                            url_args={'sha1_git': revision},
                            query_params={'per_page': per_page})
    next_page_url = reverse('browse-revision-log',
                            url_args={'sha1_git': revision},
                            query_params={'offset': 2 * per_page,
                                          'per_page': per_page})

    nb_log_entries = len(revision_log_sorted) - per_page
    if nb_log_entries > per_page:
        nb_log_entries = per_page

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/revision-log.html')
    assert_contains(resp, '<tr class="swh-revision-log-entry',
                    count=nb_log_entries)

    assert_contains(resp, '<a class="page-link" href="%s">Newer</a>' %
                    escape(prev_page_url))

    if len(revision_log_sorted) > 2 * per_page:
        assert_contains(resp, '<a class="page-link" href="%s">Older</a>' %  # noqa
                            escape(next_page_url))

    if len(revision_log_sorted) <= 2 * per_page:
        return

    resp = client.get(next_page_url)

    prev_page_url = reverse('browse-revision-log',
                            url_args={'sha1_git': revision},
                            query_params={'offset': per_page,
                                          'per_page': per_page})
    next_page_url = reverse('browse-revision-log',
                            url_args={'sha1_git': revision},
                            query_params={'offset': 3 * per_page,
                                          'per_page': per_page})

    nb_log_entries = len(revision_log_sorted) - 2 * per_page
    if nb_log_entries > per_page:
        nb_log_entries = per_page

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/revision-log.html')
    assert_contains(resp, '<tr class="swh-revision-log-entry',
                    count=nb_log_entries)
    assert_contains(resp, '<a class="page-link" href="%s">Newer</a>' %
                    escape(prev_page_url))

    if len(revision_log_sorted) > 3 * per_page:
        assert_contains(resp, '<a class="page-link" href="%s">Older</a>' %  # noqa
                            escape(next_page_url))


@given(revision(), unknown_revision(), new_origin())
def test_revision_request_errors(client, revision, unknown_revision,
                                 new_origin):
    url = reverse('browse-revision',
                  url_args={'sha1_git': unknown_revision})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp,
                    'Revision with sha1_git %s not found' %
                    unknown_revision, status_code=404)

    url = reverse('browse-revision',
                  url_args={'sha1_git': revision},
                  query_params={'origin': new_origin['url']})

    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, 'the origin mentioned in your request'
                    ' appears broken', status_code=404)


@given(revision())
def test_revision_uppercase(client, revision):
    url = reverse('browse-revision-uppercase-checksum',
                  url_args={'sha1_git': revision.upper()})

    resp = client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('browse-revision',
                           url_args={'sha1_git': revision})

    assert resp['location'] == redirect_url



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.conf.urls import url
from django.core.cache import caches
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.cache import never_cache
from django.views.decorators.clickjacking import xframe_options_exempt

from swh.web.common import service
from swh.web.common.exc import handle_view_exception
from swh.web.config import get_config


# Current coverage list of the archive
# TODO: Retrieve that list dynamically instead of hardcoding it
_code_providers = [
    {
        'provider_id': 'bitbucket',
        'provider_url': 'https://bitbucket.org/',
        'provider_logo': 'img/logos/bitbucket.png',
        'provider_info': 'public repositories from Bitbucket '
                         '(continuously archived)',
        'origin_url_regexp': '^https://bitbucket.org/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'debian',
        'provider_url': 'https://www.debian.org/',
        'provider_logo': 'img/logos/debian.png',
        'provider_info': 'source packages from the Debian distribution '
                         '(continuously archived)',
        'origin_url_regexp': '^deb://',
        'origin_types': 'packages',
    },
    {
        'provider_id': 'framagit',
        'provider_url': 'https://framagit.org/',
        'provider_logo': 'img/logos/framagit.png',
        'provider_info': 'public repositories from Framagit '
                         '(continuously archived)',
        'origin_url_regexp': '^https://framagit.org/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'github',
        'provider_url': 'https://github.com',
        'provider_logo': 'img/logos/github.png',
        'provider_info': 'public repositories from GitHub '
                         '(continuously archived)',
        'origin_url_regexp': '^https://github.com/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'gitlab',
        'provider_url': 'https://gitlab.com',
        'provider_logo': 'img/logos/gitlab.svg',
        'provider_info': 'public repositories from GitLab '
                         '(continuously archived)',
        'origin_url_regexp': '^https://gitlab.com/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'gitorious',
        'provider_url': 'https://gitorious.org/',
        'provider_logo': 'img/logos/gitorious.png',
        'provider_info': 'public repositories from the former Gitorious code '
                         'hosting service',
        'origin_url_regexp': '^https://gitorious.org/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'googlecode',
        'provider_url': 'https://code.google.com/archive/',
        'provider_logo': 'img/logos/googlecode.png',
        'provider_info': 'public repositories from the former Google Code '
                         'project hosting service',
        'origin_url_regexp': '^http.*.googlecode.com/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'gnu',
        'provider_url': 'https://www.gnu.org',
        'provider_logo': 'img/logos/gnu.png',
        'provider_info': 'releases from the GNU project (as of August 2015)',
        'origin_url_regexp': '^rsync://ftp.gnu.org/',
        'origin_types': 'releases',
    },
    {
        'provider_id': 'hal',
        'provider_url': 'https://hal.archives-ouvertes.fr/',
        'provider_logo': 'img/logos/hal.png',
        'provider_info': 'scientific software source code deposited in the '
                         'open archive HAL',
        'origin_url_regexp': '^https://hal.archives-ouvertes.fr/',
        'origin_types': 'deposits',

    },
    {
        'provider_id': 'inria',
        'provider_url': 'https://gitlab.inria.fr',
        'provider_logo': 'img/logos/inria.jpg',
        'provider_info': 'public repositories from Inria GitLab '
                         '(continuously archived)',
        'origin_url_regexp': '^https://gitlab.inria.fr/',
        'origin_types': 'repositories',
    },
    {
        'provider_id': 'npm',
        'provider_url': 'https://www.npmjs.com/',
        'provider_logo': 'img/logos/npm.png',
        'provider_info': 'public packages from the package registry for '
                         'javascript (continuously archived)',
        'origin_url_regexp': '^https://www.npmjs.com/',
        'origin_types': 'packages',
    },
    {
        'provider_id': 'pypi',
        'provider_url': 'https://pypi.org',
        'provider_logo': 'img/logos/pypi.svg',
        'provider_info': 'source packages from the Python Packaging Index '
                         '(continuously archived)',
        'origin_url_regexp': '^https://pypi.org/',
        'origin_types': 'packages',
    },
]


@xframe_options_exempt
def _swh_coverage(request):
    count_origins = get_config()['coverage_count_origins']
    return render(request, 'misc/coverage.html',
                  {'providers': _code_providers,
                   'count_origins': count_origins})


@never_cache
def _swh_coverage_count(request):
    """Internal browse endpoint to count the number of origins associated
    to each code provider declared in the archive coverage list.
    As this operation takes some times, we execute it once per day and
    cache its results to database. The cached origin counts are then served.
    Cache management is handled in the implementation to avoid sending
    the same count query twice to the storage database.
    """
    try:
        cache = caches['db_cache']
        results = []
        for code_provider in _code_providers:
            provider_id = code_provider['provider_id']
            url_regexp = code_provider['origin_url_regexp']
            cache_key = '%s_origins_count' % provider_id
            prev_cache_key = '%s_origins_prev_count' % provider_id
            # get cached origin count
            origin_count = cache.get(cache_key, -2)
            # cache entry has expired or does not exist
            if origin_count == -2:
                # mark the origin count as processing
                cache.set(cache_key, -1, timeout=10*60)
                # execute long count query
                origin_count = service.storage.origin_count(url_regexp,
                                                            regexp=True)
                # cache count result
                cache.set(cache_key, origin_count, timeout=24*60*60)
                cache.set(prev_cache_key, origin_count, timeout=None)
            # origin count is currently processing
            elif origin_count == -1:
                # return previous count if it exists
                origin_count = cache.get(prev_cache_key, -1)
            results.append({
                'provider_id': provider_id,
                'origin_count': origin_count,
                'origin_types': code_provider['origin_types']
            })
        results = json.dumps(results)
    except Exception as exc:
        return handle_view_exception(request, exc, html_response=False)

    return HttpResponse(results, content_type='application/json')


urlpatterns = [
    url(r'^coverage/$', _swh_coverage, name='swh-coverage'),
    url(r'^coverage/count/$', _swh_coverage_count, name='swh-coverage-count'),
]



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random
import re

import swh.web.browse.utils

from django.utils.html import escape

from hypothesis import given

from swh.model.hashutil import hash_to_bytes
from swh.web.browse.utils import process_snapshot_branches
from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import (
    reverse, gen_path_info, format_utc_iso_date,
    parse_timestamp, get_swh_persistent_id
)
from swh.web.tests.data import get_content
from swh.web.tests.django_asserts import assert_contains, assert_template_used
from swh.web.tests.strategies import (
    origin, origin_with_multiple_visits, new_origin,
    new_snapshot, visit_dates, revisions, origin_with_releases
)


@given(origin_with_multiple_visits())
def test_origin_visits_browse(client, archive_data, origin):
    url = reverse('browse-origin-visits',
                  url_args={'origin_url': origin['url']})
    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/origin-visits.html')

    url = reverse('browse-origin-visits',
                  url_args={'origin_url': origin['url']})
    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/origin-visits.html')

    visits = archive_data.origin_visit_get(origin['url'])

    for v in visits:
        vdate = format_utc_iso_date(v['date'], '%Y-%m-%dT%H:%M:%SZ')
        browse_dir_url = reverse('browse-origin-directory',
                                 url_args={'origin_url': origin['url'],
                                           'timestamp': vdate})
        assert_contains(resp, browse_dir_url)


@given(origin_with_multiple_visits())
def test_origin_content_view(client, archive_data, origin):
    origin_visits = archive_data.origin_visit_get(origin['url'])

    def _get_archive_data(visit_idx):
        snapshot = archive_data.snapshot_get(
            origin_visits[visit_idx]['snapshot'])
        head_rev_id = archive_data.snapshot_get_head(snapshot)
        head_rev = archive_data.revision_get(head_rev_id)
        dir_content = archive_data.directory_ls(head_rev['directory'])
        dir_files = [e for e in dir_content if e['type'] == 'file']
        dir_file = random.choice(dir_files)
        branches, releases = process_snapshot_branches(snapshot)
        return {
            'branches': branches,
            'releases': releases,
            'root_dir_sha1': head_rev['directory'],
            'content': get_content(dir_file['checksums']['sha1']),
            'visit': origin_visits[visit_idx]
        }

    tdata = _get_archive_data(-1)

    _origin_content_view_test_helper(client, origin, origin_visits,
                                     tdata['branches'],
                                     tdata['releases'],
                                     tdata['root_dir_sha1'],
                                     tdata['content'])

    _origin_content_view_test_helper(client, origin, origin_visits,
                                     tdata['branches'],
                                     tdata['releases'],
                                     tdata['root_dir_sha1'],
                                     tdata['content'],
                                     timestamp=tdata['visit']['date'])

    visit_unix_ts = parse_timestamp(tdata['visit']['date']).timestamp()
    visit_unix_ts = int(visit_unix_ts)

    _origin_content_view_test_helper(client, origin, origin_visits,
                                     tdata['branches'],
                                     tdata['releases'],
                                     tdata['root_dir_sha1'],
                                     tdata['content'],
                                     timestamp=visit_unix_ts)

    tdata = _get_archive_data(0)

    _origin_content_view_test_helper(client, origin, origin_visits,
                                     tdata['branches'],
                                     tdata['releases'],
                                     tdata['root_dir_sha1'],
                                     tdata['content'],
                                     visit_id=tdata['visit']['visit'])


@given(origin())
def test_origin_root_directory_view(client, archive_data, origin):
    origin_visits = archive_data.origin_visit_get(origin['url'])

    visit = origin_visits[-1]
    snapshot = archive_data.snapshot_get(visit['snapshot'])
    head_rev_id = archive_data.snapshot_get_head(snapshot)
    head_rev = archive_data.revision_get(head_rev_id)
    root_dir_sha1 = head_rev['directory']
    dir_content = archive_data.directory_ls(root_dir_sha1)
    branches, releases = process_snapshot_branches(snapshot)
    visit_unix_ts = parse_timestamp(visit['date']).timestamp()
    visit_unix_ts = int(visit_unix_ts)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       visit_id=visit['visit'])

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       timestamp=visit_unix_ts)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       timestamp=visit['date'])

    origin = dict(origin)
    del origin['type']

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       visit_id=visit['visit'])

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       timestamp=visit_unix_ts)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, dir_content,
                                       timestamp=visit['date'])


@given(origin())
def test_origin_sub_directory_view(client, archive_data, origin):
    origin_visits = archive_data.origin_visit_get(origin['url'])

    visit = origin_visits[-1]
    snapshot = archive_data.snapshot_get(visit['snapshot'])
    head_rev_id = archive_data.snapshot_get_head(snapshot)
    head_rev = archive_data.revision_get(head_rev_id)
    root_dir_sha1 = head_rev['directory']
    subdirs = [e for e in archive_data.directory_ls(root_dir_sha1)
               if e['type'] == 'dir']
    branches, releases = process_snapshot_branches(snapshot)
    visit_unix_ts = parse_timestamp(visit['date']).timestamp()
    visit_unix_ts = int(visit_unix_ts)

    if len(subdirs) == 0:
        return

    subdir = random.choice(subdirs)
    subdir_content = archive_data.directory_ls(subdir['target'])
    subdir_path = subdir['name']

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       visit_id=visit['visit'])

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       timestamp=visit_unix_ts)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       timestamp=visit['date'])

    origin = dict(origin)
    del origin['type']

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       visit_id=visit['visit'])

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       timestamp=visit_unix_ts)

    _origin_directory_view_test_helper(client, origin, origin_visits, branches,
                                       releases, root_dir_sha1, subdir_content,
                                       path=subdir_path,
                                       timestamp=visit['date'])


@given(origin())
def test_origin_branches(client, archive_data, origin):
    origin_visits = archive_data.origin_visit_get(origin['url'])

    visit = origin_visits[-1]
    snapshot = archive_data.snapshot_get(visit['snapshot'])
    snapshot_content = process_snapshot_branches(snapshot)

    _origin_branches_test_helper(client, origin, snapshot_content)

    origin = dict(origin)
    origin['type'] = None

    _origin_branches_test_helper(client, origin, snapshot_content)


@given(origin())
def test_origin_releases(client, archive_data, origin):
    origin_visits = archive_data.origin_visit_get(origin['url'])

    visit = origin_visits[-1]
    snapshot = archive_data.snapshot_get(visit['snapshot'])
    snapshot_content = process_snapshot_branches(snapshot)

    _origin_releases_test_helper(client, origin, snapshot_content)

    origin = dict(origin)
    origin['type'] = None

    _origin_releases_test_helper(client, origin, snapshot_content)


@given(new_origin(), new_snapshot(min_size=4, max_size=4), visit_dates(),
       revisions(min_size=3, max_size=3))
def test_origin_snapshot_null_branch(client, archive_data, new_origin,
                                     new_snapshot, visit_dates, revisions):
    snp_dict = new_snapshot.to_dict()
    new_origin = archive_data.origin_add([new_origin])[0]
    for i, branch in enumerate(snp_dict['branches'].keys()):
        if i == 0:
            snp_dict['branches'][branch] = None
        else:
            snp_dict['branches'][branch] = {
                'target_type': 'revision',
                'target': hash_to_bytes(revisions[i-1]),
            }

    archive_data.snapshot_add([snp_dict])
    visit = archive_data.origin_visit_add(
        new_origin['url'], visit_dates[0], type='git')
    archive_data.origin_visit_update(new_origin['url'], visit['visit'],
                                     status='partial',
                                     snapshot=snp_dict['id'])

    url = reverse('browse-origin-directory',
                  url_args={'origin_url': new_origin['url']})
    rv = client.get(url)
    assert rv.status_code == 200


@given(new_origin(), new_snapshot(min_size=4, max_size=4), visit_dates(),
       revisions(min_size=4, max_size=4))
def test_origin_snapshot_invalid_branch(client, archive_data, new_origin,
                                        new_snapshot, visit_dates, revisions):
    snp_dict = new_snapshot.to_dict()
    new_origin = archive_data.origin_add([new_origin])[0]
    for i, branch in enumerate(snp_dict['branches'].keys()):
        snp_dict['branches'][branch] = {
            'target_type': 'revision',
            'target': hash_to_bytes(revisions[i]),
        }

    archive_data.snapshot_add([snp_dict])
    visit = archive_data.origin_visit_add(
        new_origin['url'], visit_dates[0], type='git')
    archive_data.origin_visit_update(new_origin['url'], visit['visit'],
                                     status='full',
                                     snapshot=snp_dict['id'])

    url = reverse('browse-origin-directory',
                  url_args={'origin_url': new_origin['url']},
                  query_params={'branch': 'invalid_branch'})
    rv = client.get(url)
    assert rv.status_code == 404


def test_origin_request_errors(client, archive_data, mocker):
    mock_snapshot_service = mocker.patch(
        'swh.web.browse.views.utils.snapshot_context.service')
    mock_origin_service = mocker.patch('swh.web.browse.views.origin.service')
    mock_utils_service = mocker.patch('swh.web.browse.utils.service')
    mock_get_origin_visit_snapshot = mocker.patch(
        'swh.web.browse.utils.get_origin_visit_snapshot')
    mock_get_origin_visits = mocker.patch(
        'swh.web.common.origin_visits.get_origin_visits')
    mock_request_content = mocker.patch(
        'swh.web.browse.views.utils.snapshot_context.request_content')
    mock_origin_service.lookup_origin.side_effect = NotFoundExc(
        'origin not found')
    url = reverse('browse-origin-visits',
                  url_args={'origin_url': 'bar'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, 'origin not found', status_code=404)

    mock_origin_service.lookup_origin.side_effect = None
    mock_origin_service.lookup_origin.return_value = {'type': 'foo',
                                                      'url': 'bar',
                                                      'id': 457}
    mock_get_origin_visits.return_value = []
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': 'bar'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, "No visit", status_code=404)

    mock_get_origin_visits.return_value = [{'visit': 1}]
    mock_get_origin_visit_snapshot.side_effect = NotFoundExc('visit not found')
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': 'bar'},
                  query_params={'visit_id': 2})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert re.search('Visit.*not found', resp.content.decode('utf-8'))

    mock_get_origin_visits.return_value = [{
        'date': '2015-09-26T09:30:52.373449+00:00',
        'metadata': {},
        'origin': 457,
        'snapshot': 'bdaf9ac436488a8c6cda927a0f44e172934d3f65',
        'status': 'full',
        'visit': 1
    }]
    mock_get_origin_visit_snapshot.side_effect = None
    mock_get_origin_visit_snapshot.return_value = (
        [{'directory': 'ae59ceecf46367e8e4ad800e231fc76adc3afffb',
            'name': 'HEAD',
            'revision': '7bc08e1aa0b08cb23e18715a32aa38517ad34672',
            'date': '04 May 2017, 13:27 UTC',
            'message': ''}],
        []
    )
    mock_utils_service.lookup_snapshot_sizes.return_value = {
        'revision': 1,
        'release': 0
    }
    mock_lookup_directory = mock_utils_service.lookup_directory
    mock_lookup_directory.side_effect = NotFoundExc('Directory not found')
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': 'bar'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, 'Directory not found', status_code=404)

    mock_origin_service.lookup_origin.side_effect = None
    mock_origin_service.lookup_origin.return_value = {'type': 'foo',
                                                      'url': 'bar',
                                                      'id': 457}
    mock_get_origin_visits.return_value = []
    url = reverse('browse-origin-content',
                  url_args={'origin_url': 'bar',
                            'path': 'foo'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, "No visit", status_code=404)

    mock_get_origin_visits.return_value = [{'visit': 1}]
    mock_get_origin_visit_snapshot.side_effect = NotFoundExc('visit not found')
    url = reverse('browse-origin-content',
                  url_args={'origin_url': 'bar',
                            'path': 'foo'},
                  query_params={'visit_id': 2})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert re.search('Visit.*not found', resp.content.decode('utf-8'))

    mock_get_origin_visits.return_value = [{
        'date': '2015-09-26T09:30:52.373449+00:00',
        'metadata': {},
        'origin': 457,
        'snapshot': 'bdaf9ac436488a8c6cda927a0f44e172934d3f65',
        'status': 'full',
        'type': 'git',
        'visit': 1
    }]
    mock_get_origin_visit_snapshot.side_effect = None
    mock_get_origin_visit_snapshot.return_value = ([], [])
    mock_utils_service.lookup_snapshot_sizes.return_value = {
        'revision': 0,
        'release': 0
    }
    mock_utils_service.lookup_origin.return_value = {'type': 'foo',
                                                     'url': 'bar',
                                                     'id': 457}
    url = reverse('browse-origin-content',
                  url_args={'origin_url': 'bar',
                            'path': 'baz'})
    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')
    assert re.search('snapshot.*is empty', resp.content.decode('utf-8'))

    mock_get_origin_visit_snapshot.return_value = (
        [{'directory': 'ae59ceecf46367e8e4ad800e231fc76adc3afffb',
            'name': 'HEAD',
            'revision': '7bc08e1aa0b08cb23e18715a32aa38517ad34672',
            'date': '04 May 2017, 13:27 UTC',
            'message': ''}],
        []
    )
    mock_utils_service.lookup_snapshot_sizes.return_value = {
        'revision': 1,
        'release': 0
    }
    mock_snapshot_service.lookup_directory_with_path.return_value = {
        'target': '5ecd9f37b7a2d2e9980d201acd6286116f2ba1f1'
    }
    mock_request_content.side_effect = NotFoundExc('Content not found')
    url = reverse('browse-origin-content',
                  url_args={'origin_url': 'bar',
                            'path': 'baz'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, 'Content not found', status_code=404)

    mock_get_snapshot_context = mocker.patch(
        'swh.web.browse.views.utils.snapshot_context.get_snapshot_context')

    mock_get_snapshot_context.side_effect = NotFoundExc('Snapshot not found')
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': 'bar'})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')
    assert_contains(resp, 'Snapshot not found', status_code=404)


def test_origin_empty_snapshot(client, mocker):
    mock_utils_service = mocker.patch('swh.web.browse.utils.service')
    mock_get_origin_visit_snapshot = mocker.patch(
        'swh.web.browse.utils.get_origin_visit_snapshot')
    mock_get_origin_visits = mocker.patch(
        'swh.web.common.origin_visits.get_origin_visits')
    mock_get_origin_visits.return_value = [{
        'date': '2015-09-26T09:30:52.373449+00:00',
        'metadata': {},
        'origin': 457,
        'snapshot': 'bdaf9ac436488a8c6cda927a0f44e172934d3f65',
        'status': 'full',
        'type': 'git',
        'visit': 1
    }]
    mock_get_origin_visit_snapshot.return_value = ([], [])
    mock_utils_service.lookup_snapshot_sizes.return_value = {
        'revision': 0,
        'release': 0
    }
    mock_utils_service.lookup_origin.return_value = {
        'id': 457,
        'url': 'https://github.com/foo/bar'
    }
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': 'bar'})
    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/directory.html')
    resp_content = resp.content.decode('utf-8')
    assert re.search('snapshot.*is empty', resp_content)
    assert not re.search('swh-tr-link', resp_content)


@given(origin_with_releases())
def test_origin_release_browse(client, archive_data, origin):
    # for swh.web.browse.utils.get_snapshot_content to only return one branch
    snapshot_max_size = swh.web.browse.utils.snapshot_content_max_size
    swh.web.browse.utils.snapshot_content_max_size = 1
    try:
        snapshot = archive_data.snapshot_get_latest(origin['url'])
        release = [b for b in snapshot['branches'].values()
                   if b['target_type'] == 'release'][-1]
        release_data = archive_data.release_get(release['target'])
        url = reverse('browse-origin-directory',
                      url_args={'origin_url': origin['url']},
                      query_params={'release': release_data['name']})

        resp = client.get(url)
        assert resp.status_code == 200
        assert_contains(resp, release_data['name'])
        assert_contains(resp, release['target'])
    finally:
        swh.web.browse.utils.snapshot_content_max_size = snapshot_max_size


@given(origin_with_releases())
def test_origin_release_browse_not_found(client, archive_data, origin):

    invalid_release_name = 'swh-foo-bar'
    url = reverse('browse-origin-directory',
                  url_args={'origin_url': origin['url']},
                  query_params={'release': invalid_release_name})

    resp = client.get(url)
    assert resp.status_code == 404
    assert re.search(f'Release {invalid_release_name}.*not found',
                     resp.content.decode('utf-8'))


def _origin_content_view_test_helper(client, origin_info, origin_visits,
                                     origin_branches, origin_releases,
                                     root_dir_sha1, content,
                                     visit_id=None, timestamp=None):
    content_path = '/'.join(content['path'].split('/')[1:])

    url_args = {'origin_url': origin_info['url'],
                'path': content_path}

    if not visit_id:
        visit_id = origin_visits[-1]['visit']

    query_params = {}

    if timestamp:
        url_args['timestamp'] = timestamp

    if visit_id:
        query_params['visit_id'] = visit_id

    url = reverse('browse-origin-content',
                  url_args=url_args,
                  query_params=query_params)

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    assert type(content['data']) == str

    assert_contains(resp, '<code class="%s">' %
                    content['hljs_language'])
    assert_contains(resp, escape(content['data']))

    split_path = content_path.split('/')

    filename = split_path[-1]
    path = content_path.replace(filename, '')[:-1]

    path_info = gen_path_info(path)

    del url_args['path']

    if timestamp:
        url_args['timestamp'] = format_utc_iso_date(
            parse_timestamp(timestamp).isoformat(), '%Y-%m-%dT%H:%M:%S')

    root_dir_url = reverse('browse-origin-directory',
                           url_args=url_args,
                           query_params=query_params)

    assert_contains(resp, '<li class="swh-path">',
                    count=len(path_info)+1)

    assert_contains(resp, '<a href="%s">%s</a>' %
                    (root_dir_url, root_dir_sha1[:7]))

    for p in path_info:
        url_args['path'] = p['path']
        dir_url = reverse('browse-origin-directory',
                          url_args=url_args,
                          query_params=query_params)
        assert_contains(resp, '<a href="%s">%s</a>' %
                        (dir_url, p['name']))

    assert_contains(resp, '<li>%s</li>' % filename)

    query_string = 'sha1_git:' + content['sha1_git']

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': query_string},
                      query_params={'filename': filename})
    assert_contains(resp, url_raw)

    if 'args' in url_args:
        del url_args['path']

    origin_branches_url = reverse('browse-origin-branches',
                                  url_args=url_args,
                                  query_params=query_params)

    assert_contains(resp, '<a href="%s">Branches (%s)</a>' %
                    (origin_branches_url, len(origin_branches)))

    origin_releases_url = reverse('browse-origin-releases',
                                  url_args=url_args,
                                  query_params=query_params)

    assert_contains(resp, '<a href="%s">Releases (%s)</a>' %
                    (origin_releases_url, len(origin_releases)))

    assert_contains(resp, '<li class="swh-branch">',
                    count=len(origin_branches))

    url_args['path'] = content_path

    for branch in origin_branches:
        query_params['branch'] = branch['name']
        root_dir_branch_url = reverse('browse-origin-content',
                                      url_args=url_args,
                                      query_params=query_params)

        assert_contains(resp, '<a href="%s">' % root_dir_branch_url)

    assert_contains(resp, '<li class="swh-release">',
                    count=len(origin_releases))

    query_params['branch'] = None
    for release in origin_releases:
        query_params['release'] = release['name']
        root_dir_release_url = reverse('browse-origin-content',
                                       url_args=url_args,
                                       query_params=query_params)

        assert_contains(resp, '<a href="%s">' % root_dir_release_url)

    url = reverse('browse-origin-content',
                  url_args=url_args,
                  query_params=query_params)

    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    swh_cnt_id = get_swh_persistent_id('content', content['sha1_git'])
    swh_cnt_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_cnt_id})
    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)

    assert_contains(resp, 'swh-take-new-snapshot')


def _origin_directory_view_test_helper(client, origin_info, origin_visits,
                                       origin_branches, origin_releases,
                                       root_directory_sha1, directory_entries,
                                       visit_id=None, timestamp=None,
                                       path=None):
    dirs = [e for e in directory_entries
            if e['type'] in ('dir', 'rev')]
    files = [e for e in directory_entries
             if e['type'] == 'file']

    if not visit_id:
        visit_id = origin_visits[-1]['visit']

    url_args = {'origin_url': origin_info['url']}

    query_params = {}

    if timestamp:
        url_args['timestamp'] = timestamp
    else:
        query_params['visit_id'] = visit_id

    if path:
        url_args['path'] = path

    url = reverse('browse-origin-directory',
                  url_args=url_args,
                  query_params=query_params)

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/directory.html')

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/directory.html')

    assert_contains(resp, '<td class="swh-directory">',
                    count=len(dirs))
    assert_contains(resp, '<td class="swh-content">',
                    count=len(files))

    if timestamp:
        url_args['timestamp'] = format_utc_iso_date(
            parse_timestamp(timestamp).isoformat(), '%Y-%m-%dT%H:%M:%S')

    for d in dirs:
        if d['type'] == 'rev':
            dir_url = reverse('browse-revision',
                              url_args={'sha1_git': d['target']})
        else:
            dir_path = d['name']
            if path:
                dir_path = "%s/%s" % (path, d['name'])
            dir_url_args = dict(url_args)
            dir_url_args['path'] = dir_path
            dir_url = reverse('browse-origin-directory',
                              url_args=dir_url_args,
                              query_params=query_params)
        assert_contains(resp, dir_url)

    for f in files:
        file_path = f['name']
        if path:
            file_path = "%s/%s" % (path, f['name'])
        file_url_args = dict(url_args)
        file_url_args['path'] = file_path
        file_url = reverse('browse-origin-content',
                           url_args=file_url_args,
                           query_params=query_params)
        assert_contains(resp, file_url)

    if 'path' in url_args:
        del url_args['path']

    root_dir_branch_url = reverse('browse-origin-directory',
                                  url_args=url_args,
                                  query_params=query_params)

    nb_bc_paths = 1
    if path:
        nb_bc_paths = len(path.split('/')) + 1

    assert_contains(resp, '<li class="swh-path">', count=nb_bc_paths)
    assert_contains(resp, '<a href="%s">%s</a>' %
                    (root_dir_branch_url,
                     root_directory_sha1[:7]))

    origin_branches_url = reverse('browse-origin-branches',
                                  url_args=url_args,
                                  query_params=query_params)

    assert_contains(resp, '<a href="%s">Branches (%s)</a>' %
                    (origin_branches_url, len(origin_branches)))

    origin_releases_url = reverse('browse-origin-releases',
                                  url_args=url_args,
                                  query_params=query_params)

    nb_releases = len(origin_releases)
    if nb_releases > 0:
        assert_contains(resp, '<a href="%s">Releases (%s)</a>' %
                        (origin_releases_url, nb_releases))

    if path:
        url_args['path'] = path

    assert_contains(resp, '<li class="swh-branch">',
                    count=len(origin_branches))

    for branch in origin_branches:
        query_params['branch'] = branch['name']
        root_dir_branch_url = reverse('browse-origin-directory',
                                      url_args=url_args,
                                      query_params=query_params)

        assert_contains(resp, '<a href="%s">' % root_dir_branch_url)

    assert_contains(resp, '<li class="swh-release">',
                    count=len(origin_releases))

    query_params['branch'] = None
    for release in origin_releases:
        query_params['release'] = release['name']
        root_dir_release_url = reverse('browse-origin-directory',
                                       url_args=url_args,
                                       query_params=query_params)

        assert_contains(resp, '<a href="%s">' % root_dir_release_url)

    assert_contains(resp, 'vault-cook-directory')
    assert_contains(resp, 'vault-cook-revision')

    swh_dir_id = get_swh_persistent_id('directory', directory_entries[0]['dir_id'])  # noqa
    swh_dir_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_dir_id})
    assert_contains(resp, swh_dir_id)
    assert_contains(resp, swh_dir_id_url)

    assert_contains(resp, 'swh-take-new-snapshot')


def _origin_branches_test_helper(client, origin_info, origin_snapshot):
    url_args = {'origin_url': origin_info['url']}

    url = reverse('browse-origin-branches',
                  url_args=url_args)

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/branches.html')

    origin_branches = origin_snapshot[0]
    origin_releases = origin_snapshot[1]

    origin_branches_url = reverse('browse-origin-branches',
                                  url_args=url_args)

    assert_contains(resp, '<a href="%s">Branches (%s)</a>' %
                    (origin_branches_url, len(origin_branches)))

    origin_releases_url = reverse('browse-origin-releases',
                                  url_args=url_args)

    nb_releases = len(origin_releases)
    if nb_releases > 0:
        assert_contains(resp, '<a href="%s">Releases (%s)</a>' %
                        (origin_releases_url, nb_releases))

    assert_contains(resp, '<tr class="swh-branch-entry',
                    count=len(origin_branches))

    for branch in origin_branches:
        browse_branch_url = reverse(
            'browse-origin-directory',
            url_args={'origin_url': origin_info['url']},
            query_params={'branch': branch['name']})
        assert_contains(resp, '<a href="%s">' %
                        escape(browse_branch_url))

        browse_revision_url = reverse(
            'browse-revision',
            url_args={'sha1_git': branch['revision']},
            query_params={'origin': origin_info['url']})
        assert_contains(resp, '<a href="%s">' %
                        escape(browse_revision_url))


def _origin_releases_test_helper(client, origin_info, origin_snapshot):
    url_args = {'origin_url': origin_info['url']}

    url = reverse('browse-origin-releases',
                  url_args=url_args)

    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/releases.html')

    origin_branches = origin_snapshot[0]
    origin_releases = origin_snapshot[1]

    origin_branches_url = reverse('browse-origin-branches',
                                  url_args=url_args)

    assert_contains(resp, '<a href="%s">Branches (%s)</a>' %
                    (origin_branches_url, len(origin_branches)))

    origin_releases_url = reverse('browse-origin-releases',
                                  url_args=url_args)

    nb_releases = len(origin_releases)
    if nb_releases > 0:
        assert_contains(resp, '<a href="%s">Releases (%s)</a>' %
                        (origin_releases_url, nb_releases))

    assert_contains(resp, '<tr class="swh-release-entry',
                    count=nb_releases)

    for release in origin_releases:
        browse_release_url = reverse(
            'browse-release',
            url_args={'sha1_git': release['id']},
            query_params={'origin': origin_info['url']})
        browse_revision_url = reverse(
            'browse-revision',
            url_args={'sha1_git': release['target']},
            query_params={'origin': origin_info['url']})

        assert_contains(resp, '<a href="%s">' %
                        escape(browse_release_url))
        assert_contains(resp, '<a href="%s">' %
                        escape(browse_revision_url))



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from unittest.mock import patch

import swh.web.gunicorn_config as gunicorn_config


def test_post_fork_default():
    with patch('sentry_sdk.init') as sentry_sdk_init:
        gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_not_called()


def test_post_fork_with_dsn_env():
    django_integration = object()  # unique object to check for equality
    with patch('swh.web.gunicorn_config.DjangoIntegration',
               new=lambda: django_integration):
        with patch('sentry_sdk.init') as sentry_sdk_init:
            with patch.dict(os.environ, {'SWH_SENTRY_DSN': 'test_dsn'}):
                gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_called_once_with(
        dsn='test_dsn',
        integrations=[django_integration],
        debug=False,
        release=None,
    )


def test_post_fork_debug():
    django_integration = object()  # unique object to check for equality
    with patch('swh.web.gunicorn_config.DjangoIntegration',
               new=lambda: django_integration):
        with patch('sentry_sdk.init') as sentry_sdk_init:
            with patch.dict(os.environ, {'SWH_SENTRY_DSN': 'test_dsn',
                                         'SWH_SENTRY_DEBUG': '1'}):
                gunicorn_config.post_fork(None, None)

    sentry_sdk_init.assert_called_once_with(
        dsn='test_dsn',
        integrations=[django_integration],
        debug=True,
        release=None,
    )



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# flake8: noqa

from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('swh.web.common', '0003_saveoriginrequest_loading_task_status'),
    ]

    operations = [
        migrations.AlterField(
            model_name='saveoriginrequest',
            name='loading_task_status',
            field=models.TextField(choices=[('not created', 'not created'), ('not yet scheduled', 'not yet scheduled'), ('scheduled', 'scheduled'), ('succeed', 'succeed'), ('failed', 'failed'), ('running', 'running')], default='not created'),
        ),
    ]



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import math

from django.core.cache import cache

from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import parse_timestamp


def get_origin_visits(origin_info):
    """Function that returns the list of visits for a swh origin.
    That list is put in cache in order to speedup the navigation
    in the swh web browse ui.

    Args:
        origin_info (dict): dict describing the origin to fetch visits from

    Returns:
        list: A list of dict describing the origin visits with the
        following keys:

            * **date**: UTC visit date in ISO format,
            * **origin**: the origin id
            * **status**: the visit status, either **full**, **partial**
              or **ongoing**
            * **visit**: the visit id

    Raises:
        NotFoundExc: if the origin is not found
    """

    from swh.web.common import service

    if 'url' in origin_info:
        origin_url = origin_info['url']
    else:
        origin_url = service.lookup_origin(origin_info)['url']

    cache_entry_id = 'origin_visits_%s' % origin_url
    cache_entry = cache.get(cache_entry_id)

    if cache_entry:
        last_visit = cache_entry[-1]['visit']
        new_visits = list(service.lookup_origin_visits(origin_url,
                                                       last_visit=last_visit))
        if not new_visits:
            last_snp = service.lookup_latest_origin_snapshot(origin_url)
            if not last_snp or last_snp['id'] == cache_entry[-1]['snapshot']:
                return cache_entry

    origin_visits = []

    per_page = service.MAX_LIMIT
    last_visit = None
    while 1:
        visits = list(service.lookup_origin_visits(origin_url,
                                                   last_visit=last_visit,
                                                   per_page=per_page))
        origin_visits += visits
        if len(visits) < per_page:
            break
        else:
            if not last_visit:
                last_visit = per_page
            else:
                last_visit += per_page

    def _visit_sort_key(visit):
        ts = parse_timestamp(visit['date']).timestamp()
        return ts + (float(visit['visit']) / 10e3)

    for v in origin_visits:
        if 'metadata' in v:
            del v['metadata']
    origin_visits = [dict(t) for t in set([tuple(d.items())
                                           for d in origin_visits])]
    origin_visits = sorted(origin_visits, key=lambda v: _visit_sort_key(v))

    cache.set(cache_entry_id, origin_visits)

    return origin_visits


def get_origin_visit(origin_info, visit_ts=None, visit_id=None,
                     snapshot_id=None):
    """Function that returns information about a visit for
    a given origin.
    The visit is retrieved from a provided timestamp.
    The closest visit from that timestamp is selected.

    Args:
        origin_info (dict): a dict filled with origin information
            (id, url, type)
        visit_ts (int or str): an ISO date string or Unix timestamp to parse

    Returns:
        A dict containing the visit info as described below::

            {'origin': 2,
             'date': '2017-10-08T11:54:25.582463+00:00',
             'metadata': {},
             'visit': 25,
             'status': 'full'}

    """
    visits = get_origin_visits(origin_info)

    if not visits:
        if 'url' in origin_info:
            message = ('No visit associated to origin with'
                       ' url %s!' % origin_info['url'])
        else:
            message = ('No visit associated to origin with'
                       ' id %s!' % origin_info['id'])
        raise NotFoundExc(message)

    if snapshot_id:
        visit = [v for v in visits if v['snapshot'] == snapshot_id]
        if len(visit) == 0:
            if 'type' in origin_info and 'url' in origin_info:
                message = ('Visit for snapshot with id %s for origin with type'
                           ' url %s not found!' %
                           (snapshot_id, origin_info['url']))
            else:
                message = ('Visit for snapshot with id %s for origin with'
                           ' id %s not found!' %
                           (snapshot_id, origin_info['id']))
            raise NotFoundExc(message)
        return visit[0]

    if visit_id:
        visit = [v for v in visits if v['visit'] == int(visit_id)]
        if len(visit) == 0:
            if 'type' in origin_info and 'url' in origin_info:
                message = ('Visit with id %s for origin with'
                           ' and url %s not found!' %
                           (visit_id, origin_info['url']))
            else:
                message = ('Visit with id %s for origin with id %s'
                           ' not found!' % (visit_id, origin_info['id']))
            raise NotFoundExc(message)
        return visit[0]

    if not visit_ts:
        # returns the latest full visit when no timestamp is provided
        for v in reversed(visits):
            if v['status'] == 'full':
                return v
        return visits[-1]

    target_visit_ts = math.floor(parse_timestamp(visit_ts).timestamp())

    # Find the visit with date closest to the target (in absolute value)
    (abs_time_delta, visit_idx) = min(
        ((math.floor(parse_timestamp(visit['date']).timestamp()), i)
         for (i, visit) in enumerate(visits)),
        key=lambda ts_and_i: abs(ts_and_i[0] - target_visit_ts))

    if visit_idx is not None:
        visit = visits[visit_idx]
        # If multiple visits have the same date, select the one with
        # the largest id.
        while visit_idx < len(visits) - 1 and \
                visit['date'] == visits[visit_idx+1]['date']:
            visit_idx = visit_idx + 1
            visit = visits[visit_idx]
        return visit
    else:
        if 'type' in origin_info and 'url' in origin_info:
            message = ('Visit with timestamp %s for origin with '
                       'and url %s not found!' %
                       (visit_ts, origin_info['url']))
        else:
            message = ('Visit with timestamp %s for origin with id %s '
                       'not found!' % (visit_ts, origin_info['id']))
        raise NotFoundExc(message)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

import requests
import sentry_sdk

from django.conf.urls import url, include
from django.contrib.staticfiles import finders
from django.http import HttpResponse
from django.shortcuts import render

from swh.web.common import service
from swh.web.config import get_config
from swh.web.misc.metrics import prometheus_metrics


def _jslicenses(request):
    jslicenses_file = finders.find('jssources/jslicenses.json')
    jslicenses_data = json.load(open(jslicenses_file))
    jslicenses_data = sorted(jslicenses_data.items(),
                             key=lambda item: item[0].split('/')[-1])
    return render(request, "misc/jslicenses.html",
                  {'jslicenses_data': jslicenses_data})


def _stat_counters(request):
    stat = service.stat_counters()
    url = get_config()['history_counters_url']
    stat_counters_history = 'null'
    if url:
        try:
            response = requests.get(url, timeout=5)
            stat_counters_history = response.text
        except Exception as exc:
            sentry_sdk.capture_exception(exc)
    json_data = '{"stat_counters": %s, "stat_counters_history": %s}' % (
        json.dumps(stat), stat_counters_history)
    return HttpResponse(json_data, content_type='application/json')


urlpatterns = [
    url(r'^', include('swh.web.misc.coverage')),
    url(r'^jslicenses/$', _jslicenses, name='jslicenses'),
    url(r'^', include('swh.web.misc.origin_save')),
    url(r'^stat_counters/', _stat_counters, name='stat-counters'),
    url(r'^', include('swh.web.misc.badges')),
    url(r'^metrics/prometheus/$', prometheus_metrics,
        name='metrics-prometheus'),
]


# when running end to end tests trough cypress, declare some extra
# endpoints to provide input data for some of those tests
if get_config()['e2e_tests_mode']:
    from swh.web.tests.data import (
        get_content_code_data_by_ext,
        get_content_other_data_by_ext,
        get_content_code_data_all_exts,
        get_content_code_data_by_filename,
        get_content_code_data_all_filenames,
     ) # noqa
    urlpatterns.append(
        url(r'^tests/data/content/code/extension/(?P<ext>.+)/$',
            get_content_code_data_by_ext,
            name='tests-content-code-extension'))
    urlpatterns.append(
        url(r'^tests/data/content/other/extension/(?P<ext>.+)/$',
            get_content_other_data_by_ext,
            name='tests-content-other-extension'))
    urlpatterns.append(url(r'^tests/data/content/code/extensions/$',
                           get_content_code_data_all_exts,
                           name='tests-content-code-extensions'))
    urlpatterns.append(
        url(r'^tests/data/content/code/filename/(?P<filename>.+)/$',
            get_content_code_data_by_filename,
            name='tests-content-code-filename'))
    urlpatterns.append(url(r'^tests/data/content/code/filenames/$',
                           get_content_code_data_all_filenames,
                           name='tests-content-code-filenames'))



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.web.api.utils import enrich_directory
from swh.web.common.utils import reverse
from swh.web.tests.data import random_sha1
from swh.web.tests.strategies import directory


@given(directory())
def test_api_directory(api_client, archive_data, directory):

    url = reverse('api-1-directory', url_args={'sha1_git': directory})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'

    dir_content = list(archive_data.directory_ls(directory))
    expected_data = list(map(enrich_directory,
                             dir_content,
                             [rv.wsgi_request] * len(dir_content)))

    assert rv.data == expected_data


def test_api_directory_not_found(api_client):
    unknown_directory_ = random_sha1()

    url = reverse('api-1-directory',
                  url_args={'sha1_git': unknown_directory_})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Directory with sha1_git %s not found' % unknown_directory_
    }


@given(directory())
def test_api_directory_with_path_found(api_client, archive_data, directory):

    directory_content = archive_data.directory_ls(directory)
    path = random.choice(directory_content)

    url = reverse('api-1-directory',
                  url_args={'sha1_git': directory, 'path': path['name']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == enrich_directory(path, rv.wsgi_request)


@given(directory())
def test_api_directory_with_path_not_found(api_client, directory):

    path = 'some/path/to/nonexistent/dir/'
    url = reverse('api-1-directory',
                  url_args={'sha1_git': directory, 'path': path})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': ('Directory entry with path %s from %s not found' %
                   (path, directory))
    }


@given(directory())
def test_api_directory_uppercase(api_client, directory):
    url = reverse('api-1-directory-uppercase-checksum',
                  url_args={'sha1_git': directory.upper()})

    resp = api_client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('api-1-directory', url_args={'sha1_git': directory})

    assert resp['location'] == redirect_url



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Django development settings for swh-web.
"""

from django.core.cache import cache

from .common import * # noqa
from .common import MIDDLEWARE

MIDDLEWARE += ['swh.web.common.middlewares.HtmlPrettifyMiddleware']

AUTH_PASSWORD_VALIDATORS = []  # disable any pwd validation mechanism

cache.clear()
